
/* sort.js
   version sam. sept. 17 15:44:06 CEST 2022
   version jeu. sept. 22 02:39:56 CEST 2022

 */

POLY8=[ 0x8E, 0x95, 0x96, 0xA6,
        0XAF, 0xB1, 0XB2, 0XB4,
        0xB8, 0xC3, 0xC6, 0XD4,
        0xE1, 0xE7, 0xF3, 0xFA ];

// Some general UI stuff:
var SIZE=1<<9, // A power of 2.
    SPEED_MS=1,
    autoid, runid, running=0,
    sinfreq, sinampl, sinslope,


///////////////////////////////////////////////////////////////

/////// Manage the circular buffer of the not-yet-merged runs:
    runbuffer=[], runs, runhead, runtail,
    rbctx, rbid,
    RUNMAX= SIZE/4, // Max possible value
    RMZOOM=(SIZE/RUNMAX)|0;
    RUNMASK=RUNMAX-1;

function set_rb(v,i,col) {
  runbuffer[i]=v;
  rbctx.fillStyle= col||"#000";
  rbctx.fillRect(i*RMZOOM, 0, RMZOOM, 20);
}

function InitRunBuffer() {
  runbuffer=[];
  runs=0;
  runhead=
  runtail= RUNMAX>>1;
  rbctx.fillStyle= "#FFF";
  rbctx.fillRect(0, 0, 512, 20);
  ShowRunState();
}

function RunHeadPush(val) {
  if (runs >= RUNMAX) {
    alert("RunHeadPush : no more room");
  }
  else {
   // Check here if the runs could be merged !
    runs++;
    set_rb(val,runhead,"#0F0");
    runhead = RUNMASK & (runhead+1);
    ShowRunState();
  }
}

function RunHeadPop() {
  if (runs <= 0) {
    alert("RunHeadPop : underflow");
  }
  else {
    runs--;
    runhead = RUNMASK & (runhead-1);
    var v=runbuffer[runhead];
    set_rb(-1,runhead,"#AFA");
    ShowRunState();
    return v;
  }
}

function RunHeadTop() {
  if (runs < 1) {
    alert("RunHeadTop : no such data");
  }
  else {
    return runbuffer[RUNMASK & (runhead-1)];
  }
}

function RunHeadTop2() {
  if (runs < 2) {
    alert("RunHeadTop2 : no such data");
  }
  else {
    return runbuffer[RUNMASK & (runhead-2)];
  }
}

// The mirrored functions:

function RunTailPush(val) {
  if (runs >= RUNMAX) {
    alert("RunTailPush : no more room");
  }
  else {
   // Check here if the runs could be merged !
    runs++;
    runtail = RUNMASK & (runtail-1);
    set_rb(val,runtail,"#00F");
    ShowRunState();
  }
}

function RunTailPop() {
  if (runs <= 0) {
    alert("RunHeadPop : underflow");
  }
  else {
    runs--;
    var v=runbuffer[runtail];
    set_rb(-1,runtail,"#AAF");
    runtail = RUNMASK & (runtail+1);
    ShowRunState();
    return v;
  }
}

function RunTailTop() {
  if (runs < 1) {
    alert("RunTailTop : no such data");
  }
  else {
    return runbuffer[runtail];
  }
}

function RunTailTop2() {
  if (runs < 2) {
    alert("RunTailTop2 : no such data");
  }
  else {
    return runbuffer[RUNMASK & (runtail+1)];
  }
}

// not used or tested yet
function RunCheckContiguous() {
  if (runs > 1) {
    var i=runtail,
        j=runs-1,
        k, len, idx, nidx;

    k=runbuffer[i];
    len= k & 0xFFFF;
    idx= k >> 16;

    do {
      nidx = idx+len;     ////////   /!\ modulo /!\
      i = RUNMASK & (i+1);
      k=runbuffer[i];
      len= k & 0xFFFF;
      idx= k >> 16;

      if (idx != nidx) {
        console.log ("There is a whole in the runs !");
        return alert("There is a whole in the runs !");
      }

      j--;
    } while (j);
  }
}


///////////////////////////////////////////////////////////////

/////// Manage the canvas/working area :
var values=[], // The main "playing field"
  valdiv, valid, valctx,
  valcount, valhead, valtail,  // Manage the 2-ended "data stack" in a circular buffer way
  VALMASK = SIZE-1;

function clr() {
  InitRunBuffer();
  values.fill(0);
  valctx.globalAlpha = 1.0;
  valctx.fillStyle="#FFF";
  valctx.fillRect(0, 0, 512, 256);
}

function setval(v,i,col) {
  values[i]=v;
  valctx.globalAlpha = 0.6;
  valctx.fillStyle="#FFF";
  valctx.fillRect(i, 0, 1, 256);
  valctx.globalAlpha = 1.0;
  valctx.fillStyle= col||"#000";
  valctx.fillRect(i, 256-v, 1, v);
}


function valshade(i) {
  valctx.globalAlpha = 0.5;
  valctx.fillStyle="#FFF";
  valctx.fillRect(i, 0, 1, 256);
  valctx.globalAlpha = 1.0;
}


function initval() {
  valcount=0;
  valhead=
  valtail= SIZE>>1; // could be 42. whatever. Let's put it in the middle for clarity of intent.
  values=[];
}

function ValHeadPush(val, col) {
  if (valcount >= SIZE) {
    alert("ValHeadPush : no more room");
  }
  else {
    valcount++;
    setval(val,valhead,col||"#0F0");
    valhead = VALMASK & (valhead+1);
  }
}

function ValHeadPop() {
  if (valcount <= 0) {
    alert("ValHeadPop : underflow");
  }
  else {
    valcount--;
    valhead = VALMASK & (valhead-1);
    return values[valhead];
  }
}

function ValHeadTop() {
  if (valcount < 1) {
    alert("ValHeadTop : no such data");
  }
  else {
    return values[VALMASK & (valhead-1)];
  }
}

// The mirrored functions:

function ValTailPush(val, col) {
  if (valcount >= SIZE) {
    alert("ValTailPush : no more room");
  }
  else {
    valcount++;
    valtail = VALMASK & (valtail-1);
    setval(val,valtail,col||"#00F");
  }
}

function ValTailPop() {
  if (valcount <= 0) {
    alert("ValHeadPop : underflow");
  }
  else {
    valcount--;
    var v=values[valtail];
    valtail = VALMASK & (valtail+1);
    return v;
  }
}

function ValTailTop() {
  if (valcount < 1) {
    alert("ValTailTop : no such data");
  }
  else {
    return values[valtail];
  }
}



//////// All the test cases
function init_flat() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(42,i);
  autostart();
}

function init_step() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(3+ (((i+3)>>1) & (255-15)), i);
  autostart();
}

function init_pets() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(5+ (((i+5)>>1) & (255-15)), 255-i);
  autostart();
}

function init_step2() {
  if (running)
    return;
  clr();
  for (i=0; i<128; i++) {
    j= 13+ (((i+3)>>1) & (255-7));
    setval(j,    i);
    setval(j,255-i);
  }
  autostart();
}

function init_comb() {
  if (running)
    return;
  clr();
  for (i=0; i<128; i++) {
    setval(200,   i*2);
    setval(20, 1+ i*2);
  }
  autostart();
}

function init_bmoc() {
  if (running)
    return;
  clr();
  for (i=0; i<128; i++) {
    setval(20,     i*2);
    setval(200, 1+ i*2);
  }
  autostart();
}

function init_asc() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(i,i);
  autostart();
}

function init_desc() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(256-i,i);
  autostart();
}

function init_arnd() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(50+(i/2)+(Math.floor(Math.random()*25)),i);
  autostart();
}

function init_rnd() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(Math.floor(Math.random()*256),i);
  autostart();
}

function init_sine(freq) {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(Math.floor( 60 + (i/3) + Math.sin((i+60)/freq)*45),i);
  autostart();
}

function init_saw() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++)
    setval(Math.floor( 50 + (i/4) + ((i*5)%45) ),i);
  autostart();
}

function init_was() {
  if (running)
    return;
  clr();
  for (i=0; i<256; i++) {
    setval(Math.floor( 180 - (i/5) + ((i*-2)%45) ),i);
  }
  autostart();
}

function init_tri() {
  if (running)
    return;
  clr();
  var j=100, k=3;
  for (i=0; i<256; i++) {
    if (j>200) k=-2;
    if (j<60) k=3;
    j += k;
    setval(j,i);
  }
  autostart();
}

function init_prng() {
  if (running)
    return;
  clr();
  var POLY=this.poly;
  var subst=13;
  var LFSR=42;
  if (!POLY)
       POLY=0xB4;

  for (i=0; i<255; i++) {
    if (LFSR==subst)
      setval(0,i);
    else
      setval(LFSR,i);

    if (LFSR & 1) { LFSR = (LFSR >> 1) ^ POLY; }
    else          { LFSR = (LFSR >> 1);        }
  }
  setval(subst,255);
  autostart();
}

function autostart() {
  if (autoid.checked)
    run();
}

function autostop() {
  runid.style.visibility="visible";
  running=0;
  ShowRunState();
}

function run() {
  if (running)
    return;
  runid.style.visibility="hidden";
  running=1;

  YAMS2(); // YAMS();
} // function run()


/////// The algorithm viewed as a sort of FSM:

// globals because must persist across calls
var streamidx, // not part of the algo itself, just to stream data
    stream=[],   // temp. copy of data to sort
    extidx,  // to be removed, used only for dev/debug
    slope,
    first,
    last,
    length,
    last_valtail,
    last_valhead;

function YAMS2() {
  // copy the visualised input data to the hypothetical "source" stream
  stream = Array.from(values); // copy values into stream
  streamidx=0;
  initval();
  InitRunBuffer();
  // init the iterator's state
  slope=0;  // init the state
  length=0;

  YAMS2_iter(); // let's go.
}

// Just an iterator over the input stream:
function YAMS2_iter() {
  // Get one new data item from the stream:
  var dat=stream[streamidx];
  length++;
  YAMS2_slope(dat);

  if (streamidx >= 255) {
    // flushing pending operations here
    process_extent();
    autostop();
  }
  else {
    streamidx++;
    setTimeout(YAMS2_iter,SPEED_MS);
  }
}

/// The part that detects the slope
function YAMS2_slope(dat) {

init: {
    if (length > 1) {
      if (slope==0) { // we don't know yet which way the slope goes
        if (dat==first) {
          break init;
        }

        last=dat; // anticipated from both branches
        if (dat > first) {  // Going up !
          slope=1;  // valtail is rewinded in process_extent()
          ValHeadPush(dat, "#00F");
          return;
        }

        // else going down:
        slope=-1;   // valhead is rewinded in process_extent()
        ValTailPush(dat, "#F00");
        return;
      } // slope==0

      reinit:{
        if (slope==1) { // already going up ?
          if (dat >= last) { // still going up
            ValHeadPush(dat, "#00F");
            last=dat;
            return;
          }
          // else: end of run => stop, rewind and restart !
          break reinit;
        }

        // else slope==-1 (already going down):
        if (last >= dat) { // still going down
          ValTailPush(dat, "#F00");  // stack_push (dat, "#F00");
          last=dat;
          return;
        }
        // else: end of run => stop, rewind and restart ! fold...
      } // "reinit" named block

      // This is where we "fold" when the "run" or "extent" ends.
      length--;
      process_extent();

      slope=0;  // re-init the state
      length=1;
      // fold!
    } // (length > 1)

    // the first operations of a run
    last_valhead=valhead; // keep track for later
    last_valtail=valtail; // and rewinding one side or another
    first=dat;
    extidx=streamidx;  // to be removed, used only for logging/debug
  } // init

  ValHeadPush(dat, "#F0F"); // Speculative write on both ends
  ValTailPush(dat, "#FF0");
  valcount--; // prevents double-counting the value
}

//////////////////// The merges


function merge2low(run1base, run1length,
                   run2base, run2length,
                   destbase) {  // Final run length will be run1length+run2length
  valshade(run1base);
  valshade(run2base);

console.log("merge2low: r1:"+run1base+"-"+run1length
   +"  r2:"+run2base+"-"+run2length+"  -> "+destbase);

  var val1=values[run1base],
      val2=values[run2base];

  run1base = VALMASK & (run1base+1);
  run2base = VALMASK & (run2base+1);

  do {
    if (val2 > val1) {
      setval(val1, destbase,"#FC0");
      destbase = VALMASK & (destbase+1);
      run1length--;
      if (run1length <= 0) { //  Run1 is exhausted, memmove with Run2
        val1=val2;
        run1base=run2base;
        run1length=run2length;
        break;
      }
      valshade(run1base);
      val1=values[run1base];
      run1base = VALMASK & (run1base+1);
    } else {
      setval(val2, destbase,"#FC0");
      destbase = VALMASK & (destbase+1);
      run2length--;
      if (run2length <= 0) { // Run2 is exhausted, memmove with Run1
       	break;
      }
      valshade(run2base);
      val2=values[run2base];
      run2base = VALMASK & (run2base+1);
    }
  } while(true);

  // final memmove
  do {
    setval(val1, destbase,"#FC0");
    destbase = VALMASK & (destbase+1);
    run1length--;
    if (run1length <= 0) {
      return;
    }
    valshade(run1base);
    val1=values[run1base];
    run1base = VALMASK & (run1base+1);
  } while (true);
}

function merge2high(run1base, run1length,
                    run2base, run2length,
                    destbase) {  // Final run length will be run1length+run2length
  var val1,
      val2,
  // point to the last element of each run
      destidx=VALMASK & (destbase+run1length+run2length-1);
  run1base = VALMASK & (run1base+run1length-1);
  run2base = VALMASK & (run2base+run2length-1);
  val1=values[run1base];
  val2=values[run2base];
  valshade(run1base);
  valshade(run2base);

console.log("merge2high: r1:"+run1base+"-"+run1length
   +"  r2:"+run2base+"-"+run2length+"  -> "+destbase);


  do {
    if (val2 <= val1) {
      setval(val1, destidx,"#096");
      destidx = VALMASK & (destidx-1);
      run1length--;
      if (run1length <= 0) {
        val1=val2;
        run1base=run2base;
        run1length=run2length;
        break;
      }
      run1base = VALMASK & (run1base-1);
      valshade(run1base);
      val1=values[run1base];
    } else {
      setval(val2, destidx,"#0FC");
      destidx = VALMASK & (destidx-1);
      run2length--;
      if (run2length <= 0) {
       	break;
      }
      run2base = VALMASK & (run2base-1);
      valshade(run2base);
      val2=values[run2base];
    }
  } while(true);

  // final memmove
  do {
    setval(val1, destidx,"#690");
    destidx = VALMASK & (destidx-1);
    run1length--;
    if (run1length <= 0) {
      return;
    }
    run1base = VALMASK & (run1base-1);
    valshade(run1base);
    val1=values[run1base];
  } while (true);
}



function merge_tails2head() {  // OK
  if (runs < 2) {
//    alert("merge_tails2head() : runs < 2");
    return;
  }

  var t, hl,
    t1i, t1l,
    t2i, t2l;

  t=RunTailPop();  t1i=t>>>16;  t1l=t&0xFFFF;
  t=RunTailPop();  t2i=t>>>16;  t2l=t&0xFFFF;
  hl = t1l+t2l;
// CHECK FREE ROOM HERE

  merge2low(t1i, t1l, t2i, t2l, valhead);
//merge2high(t1i, t1l, t2i, t2l, valhead); // should work identically
  RunHeadPush((valhead << 16)|hl);
  valtail = VALMASK & (valtail+hl);
  valhead = VALMASK & (valhead+hl);
  ShowRunState();
}


function merge_both2head() {
  if (runs < 2) {
    return;
  }

  var t, hl,
    t1i, t1l,
    h1i, h1l;

  t=RunHeadPop();  h1i=t>>>16;  h1l=t&0xFFFF;
  t=RunTailPop();  t1i=t>>>16;  t1l=t&0xFFFF;
  hl = t1l+h1l;

  merge2high(t1i, t1l, h1i, h1l, h1i);
  RunHeadPush((h1i << 16)|hl);
  valtail = VALMASK & (valtail+t1l);
  valhead = VALMASK & (valhead+t1l);
  ShowRunState();
}

/*
function merge_heads2tail() {
  if (runs < 2) {
    return;
  }


}


  // KAPUTT :

  var t, tl,
    h1i, h1l,
    h2i, h2l;

  t=RunHeadPop();  h1i=t>>>16;  h1l=t&0xFFFF;

  t=RunHeadPop();  h2i=t>>>16;  h2l=t&0xFFFF;
  tl = h1l+h2l;
// CHECK FREE ROOM HERE

  valhead = h2i;
  valtail = VALMASK & (valtail-tl);
        t = VALMASK & (valtail-1);
   merge2low(h1i, h1l, h2i, h2l, t);
//merge2high(h1i, h1l, h2i, h2l, t); // should work identically
  RunTailPush((t << 16)|tl);
}
*/

function merge_both2tail() {
}



function process_extent() {
//  console.log(" end of "+slope+" nslope, "+length+" values from "+extidx+" to "+(streamidx-1));

  if (slope < 0) {
    valhead=last_valhead; // rewind head
    console.log(" creating downslope run of length "+length+" from index "+valtail);
    RunTailPush(length + (valtail << 16));
  }
  else {
    valtail=last_valtail; // rewind tail
    console.log(" creating upslope run of length "+length+" from index "+last_valhead);
    RunHeadPush(length + (last_valhead << 16));
  }

  ShowRunState();
}

/*
on the "stack", each already sorted extent is described by:
* the "low" index
* the length
low+len-1 gives index of "high" value
They are encoded as length + (low << 16)

*/


////////////////////////////////////////////////////////
// Some UI for playing with the runs


function ShowRunState() {
  var s="Values:"+valcount
       +" - tail:"+valtail
       +" - head:"+valhead
       +"<br> Runs:"+runs
       +" - runtail:"+runtail
       +" - runhead:"+runhead;

  if (runs > 0) {
    var t,
      t1i, t1l, t2i, t2l,
      h1i, h1l, h2i, h2l;

    t=RunTailTop();
    t1i=t>>16;
    t1l=t&0xFFFF;
    t=RunHeadTop();
    h1i=t>>16;
    h1l=t&0xFFFF;

    if (runs > 2) {
      t=RunTailTop2();
      t2i=t>>16;
      t2l=t&0xFFFF;
      t=RunHeadTop2();
      h2i=t>>16;
      h2l=t&0xFFFF;
      s+= "<br>Tail:"  +t1i+" len:"+t1l
       +  "<br>Tail-1:"+t2i+" len:"+t2l+"  _\\ "+(t1l+t2l)
       +  "<br>Head-1:"+h2i+" len:"+h2l
       +  "<br>Head:"  +h1i+" len:"+h1l+"  _\\ "+(h1l+h2l)
       +  "<br>Head+tail="+(h1l+t1l);
    }
    else {
      s+="<br>Tail:"+t1i+" len:"+t1l
       + "<br>Head:"+h1i+" len:"+h1l;
    }
  }

  valdiv.innerHTML=s;
}

///////////// Mouse drawing on canvas

var canvas_drawing=0,
    coordid;

function canvas_mousedown(evt) {
  canvas_drawing=1;
  if (runs>0)
    clr();
}

function canvas_mouseup(evt) {
  if (canvas_drawing)
      canvas_drawing=0;
}

function canvas_move(evt) {
  var border=3,
    rect = valid.getBoundingClientRect(),
    x=((evt.clientX-rect.left)-border)|0,
    y=((evt.clientY-rect.top)-border)|0;
  if ((x >= 0) && (y >= 0)) {
    coordid.innerHTML= x+" : "+y;
    if (canvas_drawing && (x < 256)) {
      setval(0, x);
      setval(256-y, x);
    }
  }
}

////////////////////////////////////////////////////////

window.onload = function() {
  valdiv=document.getElementById("val");
  autoid=document.getElementById("auto");
  runid=document.getElementById("run");
  var bid=document.getElementById("blfsr"),
      broot=bid.parentNode;
  for (i in POLY8) {
    j=POLY8[i];
    var newbutton = bid.cloneNode(true);
    newbutton.poly=j;
    newbutton.onclick=init_prng;
    newbutton.innerHTML="LFSR "+i;
    broot.appendChild(newbutton);
  }
  bid.style.display="none";

  coordid=document.getElementById("coord");

  valid=document.getElementById("can");
  valctx=valid.getContext("2d");
  valctx.globalAlpha = 1.0;

  rbid=document.getElementById("crb");
  rbctx=rbid.getContext("2d");
  rbctx.globalAlpha = 1.0;

  canvas_drawing=0;
  valid.onmousedown=canvas_mousedown;
  document.onmouseup=canvas_mouseup;
  valid.onmousemove=canvas_move;
  valid.onmouseleave=function() {
    coordid.innerHTML="...";
  };
};

